try:
    import pg8000
    import sqlalchemy
    from sqlalchemy.exc import DatabaseError, ProgrammingError
    from google.cloud.alloydb.connector import Connector
except ImportError:
    AlloyDBEngine = None
    pass

from .database import get_vector_size
from ..logging import log

class AlloyDBClient:
    """
    A class to manage interactions with an AlloyDB instance.

    Example Usage:

    ```python
    client = AlloyDBClient(
        project_id="your-project-id",
        region="your-region",
        cluster_name="your-cluster-name",
        instance_name="your-instance-name",
        user="your-db-user",
        password="your-db-password"
    )

    # Create a database
    client.execute_sql("CREATE DATABASE my_database")

    # Execute other SQL statements
    client.execute_sql("CREATE TABLE my_table (id INT, name VARCHAR(50))")
    ```
    """

    def __init__(self, 
                 project_id: str, 
                 region: str, 
                 cluster_name:str, 
                 instance_name:str, 
                 user:str, 
                 password=None, 
                 db="postgres"):
        """Initializes the AlloyDB client.
         - project_id (str): GCP project ID where the AlloyDB instance resides.
         - region (str): The region where the AlloyDB instance is located.
         - cluster_name (str): The name of the AlloyDB cluster.
         - instance_name (str): The name of the AlloyDB instance.
         - user (str): The database user name.
         - password (str): The database user's password.
         - db_name (str): The name of the database.
        """
        self.connector = Connector()
        self.inst_uri = self._build_instance_uri(project_id, region, cluster_name, instance_name)
        self.engine = self._create_engine(self.inst_uri, user, password, db)

    def _build_instance_uri(self, project_id, region, cluster_name, instance_name):
        return f"projects/{project_id}/locations/{region}/clusters/{cluster_name}/instances/{instance_name}"

    def _create_engine(self, inst_uri, user, password, db):
        def getconn() -> pg8000.dbapi.Connection:
            conn = self.connector.connect(
                inst_uri,
                "pg8000",
                user=user,
                password=password,
                db=db,
                enable_iam_auth=True,
            )
            return conn

        engine = sqlalchemy.create_engine(
            "postgresql+pg8000://", 
            isolation_level="AUTOCOMMIT", 
            creator=getconn
        )
        engine.dialect.description_encoding = None
        log.info(f"Created AlloyDB engine for {inst_uri} and user: {user}")
        return engine

    def execute_sql(self, sql_statement):
        """Executes a given SQL statement with error handling.

         - sql_statement (str): The SQL statement to execute.
         - Returns: The result of the execution, if any.
        """
        sql_ = sqlalchemy.text(sql_statement)
        result = None
        with self.engine.connect() as conn:
            try:
                log.info(f"Executing SQL statement: {sql_}")
                result = conn.execute(sql_)
            except DatabaseError as e:
                if "already exists" in str(e):
                    log.warning(f"Error ignored: {str(e)}. Assuming object already exists.")
                else:
                    raise  
            finally:
                conn.close()

        return result

    @staticmethod
    def _and_or_ilike(sources, search_type="OR", operator="ILIKE"):
        unique_sources = set(sources)
        # Choose the delimiter based on the search_type argument
        delimiter = ' AND ' if search_type.upper() == "AND" else ' OR '

        # Build the conditional expressions based on the chosen delimiter
        conditions = delimiter.join(f"TRIM(source) {operator} '%{source}%'" for source in unique_sources)
        if not conditions:
            log.warning("Alloydb doc query found no like_patterns")
            return []
        
        return conditions

    def delete_sources_from_alloydb(self, sources, vector_name):
        """
        Deletes from both vectorstore and docstore
        """

        vector_length = get_vector_size(vector_name)

        conditions = self._and_or_ilike(sources, operator="=")

        if not conditions:
            log.warning("No conditions were specified, not deleting whole table!")
            return False

        query = f"""
            DELETE FROM {vector_name}_docstore
            WHERE {conditions};
            DELETE FROM {vector_name}_vectorstore_{vector_length}
            WHERE {conditions}
        """

        return self.execute_sql(query)

    def create_database(self, database_name):
        self.execute_sql(f'CREATE DATABASE "{database_name}"')

    def fetch_owners(self):
        owners = self.execute_sql('SELECT table_schema, table_name, privilege_type FROM information_schema.table_privileges')
        for row in owners:
            print(f"Schema: {row[0]}, Table: {row[1]}, Privilege: {row[2]}")
        return owners

    def create_schema(self, schema_name="public"):
        self.execute_sql(f'CREATE SCHEMA IF NOT EXISTS {schema_name};')

    def grant_schema_permissions(self, schema_name, users):
        for user in users:
            self.execute_sql(f'GRANT SELECT, INSERT, UPDATE, DELETE ON ALL TABLES IN SCHEMA {schema_name} TO "{user}";')
            self.execute_sql(f'GRANT USAGE, CREATE ON SCHEMA {schema_name} TO "{user}";')
            self.execute_sql(f'ALTER DEFAULT PRIVILEGES IN SCHEMA {schema_name} GRANT SELECT, INSERT, UPDATE, DELETE ON TABLES TO "{user}";')
            self.execute_sql(f'GRANT USAGE ON SCHEMA information_schema TO "{user}";')
            self.execute_sql(f'GRANT SELECT ON information_schema.columns TO "{user}";')
    
    def grant_table_permissions(self, table_name, users):
        for user in users:
            self.execute_sql(f'GRANT SELECT, INSERT, UPDATE, DELETE ON TABLE "{table_name}" TO "{user}";')
    
    def create_tables(self, vector_name, users):
        self.create_docstore_table(vector_name, users)
        self.create_vectorstore_table(vector_name, users)

    def create_docstore_table(self, vector_name: str, users):
        table_name = f"{vector_name}_vectorstore"
        sql = f'''
        CREATE TABLE IF NOT EXISTS "{table_name}" 
        (page_content TEXT, doc_id UUID, source TEXT, images_gsurls JSONB, chunk_metadata JSONB, langchain_metadata JSONB)
        '''
        self.execute_sql(sql)

        self.grant_table_permissions(table_name, users)

    def create_vectorstore_table(self, vector_name: str, users):
        from .database import get_vector_size
        vector_size = get_vector_size(vector_name)
        vectorstore_id = f"{vector_name}_{type}_{vector_size}"

        sql = f'''
        CREATE TABLE IF NOT EXISTS "{vectorstore_id}" (
        langchain_id UUID NOT NULL,
        content TEXT NOT NULL,
        embedding vector NOT NULL,
        source TEXT,
        langchain_metadata JSONB,
        docstore_doc_id UUID,
        eventTime TIMESTAMPTZ
        );
        '''
        self.execute_sql(sql)

        self.grant_table_permissions(vectorstore_id, users)


import uuid

def generate_uuid_from_object_id(object_id):
    # UUID namespace for example purposes; in a real application, you might choose a fixed namespace
    namespace = uuid.NAMESPACE_URL
    return uuid.uuid5(namespace, object_id)


from sunholo.agents import config_to_swagger
import yaml


# Test cases for config_to_swagger and generate_swagger functions
def test_config_to_swagger():
    swagger_yaml = config_to_swagger()
    swagger_dict = yaml.safe_load(swagger_yaml)
    
    assert 'swagger' in swagger_dict
    assert swagger_dict['swagger'] == '2.0'
    assert 'info' in swagger_dict
    assert swagger_dict['info']['title'] == 'Sunholo Multivac API - ${_BRANCH_NAME}'
    assert 'paths' in swagger_dict
    #TODO: more tests for specific agents



# https://cloud.google.com/vertex-ai/generative-ai/docs/extensions/create-extension
# https://cloud.google.com/vertex-ai/generative-ai/docs/model-reference/extension#python
from vertexai.preview import extensions
from .init import init_vertex
from ..logging import log
from ..utils.gcp_project import get_gcp_project
from ..utils.parsers import validate_extension_id

def get_extension_import_config(
        display_name: str, 
        description: str,
        api_spec_gcs: dict,
        service_account_name: dict,
        tool_use_examples: list):

    tool_use_examples = [
        {
            "extensionOperation": {
                "operationId": "say_hello",
            },
            "displayName": "Say hello in the requested language",
            "query": "Say hello in French",
            "requestParams": {
                "fields": [
                {
                    "key": "apiServicePrompt",
                    "value": {
                    "string_value": "French",
                    }
                }
                ]
            },
            "responseParams": {
                "fields": [
                {
                    "key": "apiServiceOutput",
                    "value": {
                    "string_value": "bonjour",
                    },
                }
                ],
            },
            "responseSummary": "Bonjour"
            }
    ]

    
    return {
        "displayName":  display_name,
        "description": description,
        "manifest": {
            "name": "EXTENSION_NAME_LLM",
            "description": "DESCRIPTION_LLM",
            "apiSpec": {
                "openApiGcsUri": api_spec_gcs,
            },
            "authConfig": {
                "authType": "OAUTH",
                "oauthConfig": {"service_account": service_account_name}
            }
        },
        "toolUseExamples": tool_use_examples, 
        }

# once an extension is available, call it in code here
def create_extension_instance(
        display_name: str,
        description: str,
        open_api_gcs_uri: str,
        llm_name: str=None,
        llm_description: str=None,
        runtime_config: dict=None,
        service_account: str=None,
):
    """
    Args:
    - display_name: for the human. parsed to be used as extension_name
    - description: for the human
    - open_api_gcs_uri: location on GCS where open_ai yaml spec is
    - llm_name: for the model.  If None, uses display_name
    - llm_description: for the model.  If None, uses description
    - service_account: If not specified, the Vertex AI Extension Service Agent is used to execute the extension.
    
    """
    project_id = get_gcp_project()
    extension_name = f"projects/{project_id}/locations/us-central1/extensions/{validate_extension_id(display_name)}"

    extension = extensions.Extension.create(
        extension_name=extension_name,
        display_name=display_name,
        description=description,
        runtime_config=runtime_config or None,
        manifest={
            "name": llm_name or display_name,
            "description": llm_description or description,
            "api_spec": {
                "open_api_gcs_uri": open_api_gcs_uri
            },
            "auth_config": {
                "auth_type": "GOOGLE_SERVICE_ACCOUNT_AUTH",
                "google_service_account_config": service_account or {},
            },
        },
    )
    log.info(f"Creating Vertex Extension: {extension=}")

    return extension



def create_extension_code_interpreter(
        code_artifacts_bucket=None
):

    # only us-central for now
    location = "us-central1"
    init_vertex(location=location)

    runtime_config=None
    if code_artifacts_bucket:
        runtime_config = {"codeInterpreterRuntimeConfig": 
                            {
                                "fileInputGcsBucket": code_artifacts_bucket,
                                "fileOutputGcsBucket": code_artifacts_bucket
                            }
                        }

    code_extension = create_extension_instance(
        display_name="Code Interpreter",
        description="This extension generates and executes code in the specified language",
        open_api_gcs_uri="gs://vertex-extension-public/code_interpreter.yaml",
        llm_name="code_interpreter_tool",
        llm_description="Google Code Interpreter Extension",
        runtime_config=runtime_config
        )
    log.info(f"Created code extension: {code_extension=}")

    return code_extension



""" from vertexai.preview import rag
from vertexai.preview.generative_models import GenerativeModel, Tool
import vertexai
 """
""" # Enhance generation
# Create a RAG retrieval tool
rag_retrieval_tool = Tool.from_retrieval(
    retrieval=rag.Retrieval(
        source=rag.VertexRagStore(
            rag_corpora=[rag_corpus.name],  # Currently only 1 corpus is allowed.
            similarity_top_k=3,  # Optional
        ),
    )
)
# Create a gemini-pro model instance
rag_model = GenerativeModel(
    model_name="gemini-1.0-pro-002", tools=[rag_retrieval_tool]
)

# Generate response
response = rag_model.generate_content("What is RAG and why it is helpful?")
print(response.text) """


VAC_SUBCONFIG_SCHEMA = {
    "type": "object",
    "properties": {
        "llm": {"type": "string"},
        "agent": {"type": "string"},
        "model": {"type": "string"},
        "prompt": {"type": "string"},
        "chunker": {
            "type": "object",
            "properties": {
                "type": {"type": "string"},
                "llm": {"type": "string"},
                "chunk_size": {"type": "integer"},
                "overlap": {"type": "integer"},
                "summarise": {
                    "type": "object",
                    "properties": {
                        "llm": {"type": "string"},
                        "model": {"type": "string"},
                        "threshold": {"type": "integer"},
                        "model_limit": {"type": "integer"}
                    }
                    },
            },
            "additionalProperties": False
        },
        "memory": {
            "type": "array",
            "items": {
                "type": "object",
                "patternProperties": {
                    ".*-vectorstore": {
                        "type": "object",
                        "properties": {
                            "vectorstore": {"type": "string"},
                            "self_query": {"type": "boolean"},
                            "provider": {"type": "string"},
                            "k": {"type": "integer"},
                            "vector_name": {"type": "string"},
                            "read_only": {"type": "boolean"},
                            "llm": {"type": "string"}
                        },
                        "required": ["vectorstore"]
                    }
                }
            }
        },
        "gcp_config": {
            "type": "object",
            "properties": {
                "project_id": {"type": "string"},
                "endpoint_id": {"type": "integer"},
                "location": {"type": "string"}
            }
        },
        "alloydb_config": {
            "type": "object",
            "properties": {
                "project_id": {"type": "string"},
                "region": {"type": "string"},
                "cluster": {"type": "string"},
                "instance": {"type": "string"},
                "database": {"type": "string"}
            },
            "required": ["project_id", "region", "cluster", "instance"]
        },
        "secrets": {
            "type": "array",
            "items": {"type": "string"}
        },
        "display_name": {"type": "string"},
        "avatar_url": {"type": "string"},
        "description": {"type": "string"},
        "memory_k": {"type": "integer"}
    },
    "required": ["llm", "agent"]
}


VAC_CONFIG_SCHEMA = {
    "type": "object",
    "properties": {
        "kind": {"type": "string"},
        "apiVersion": {"type": "string"},
        "gcp_config": {
            "type": "object",
            "properties": {
                "project_id": {"type": "string"},
                "location": {"type": "string"}
            },
            "required": ["project_id", "location"]
        },
        "vac": {
            "type": "object",
            "patternProperties": {
                ".*": VAC_SUBCONFIG_SCHEMA
            }
        }
    },
    "required": ["kind", "apiVersion", "vac"]
}

PROMPT_CONFIG_SCHEMA = {
    "type": "object",
    "properties": {
        "kind": {"type": "string"},
        "apiVersion": {"type": "string"},
        "prompts": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "chunk_summary": {"type": "string"},
                        "intro": {"type": "string"},
                        "template": {"type": "string"},
                        "chat_summary": {"type": "string"},
                        "summarise_known_question": {"type": "string"}
                    }
                }
            }
        }
    },
    "required": ["kind", "apiVersion", "prompts"]
}

SCHEMAS = {
    "vacConfig": VAC_CONFIG_SCHEMA,
    "promptConfig": PROMPT_CONFIG_SCHEMA
}




from ..utils import load_config_key
from ..logging import log
from ..database.alloydb import add_document_if_not_exists
from ..database.uuid import generate_uuid_from_object_id
from ..components.llm import llm_str_to_llm
from ..gcs.add_file import add_file_to_gcs, get_summary_file_name
from ..utils.parsers import remove_whitespace
from .images import upload_doc_images
from ..langfuse.callback import create_langfuse_callback

import tempfile
import traceback
import json
import os
from langchain.docstore.document import Document

from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser

def send_doc_to_docstore(docs, vector_name):

    # docs all come from the same file but got split into a list of document objects

    docstore_config = load_config_key("docstore", vector_name=vector_name, kind="vacConfig")
    if docstore_config is None:
        log.info(f"No docstore config found for {vector_name} ")
        
        return None, docs
    
    log.info(f"Docstore config: {docstore_config}")

    doc_id, big_doc, docs = create_big_doc(docs)

    for docstore in docstore_config:
        for key, value in docstore.items(): 
            log.info(f"Found memory {key}")
            type = value.get('type')
            if type == 'alloydb':
                # upload to alloydb
                log.info(f"Uploading to docstore alloydb the docs for {vector_name}")

                saved_doc_id = add_document_if_not_exists(big_doc, vector_name=vector_name)
                if saved_doc_id is not None:
                    if saved_doc_id != doc_id:
                        raise ValueError(f"Something went wrong with doc_ids: {doc_id} vs saved_doc_id: {saved_doc_id}")

            #elif docstore.get('type') == 'cloudstorage':
            else:
                log.info(f"No docstore type found for {vector_name}: {docstore}")
    
    log.info(f"Added doc to docstores: {doc_id}")
    return doc_id, docs

def create_big_doc(docs):

    if not docs:
        return None, None, None
    
    # merge docs into one document object
    big_doc = Document(page_content="", 
                        metadata={"images_gsurls": [],
                                  "chunk_metadata": []})
    doc_id = None
    for doc in docs:
        if doc_id is None:
            first_source = doc.metadata.get("objectId") or doc.metadata.get("url")
            if first_source:
                doc_id = generate_uuid_from_object_id(first_source)
        
        if doc_id is None:
            raise ValueError(f"Failed to create a doc_id from {doc.metadata}")

        doc.metadata["docstore_doc_id"] = str(doc_id)

        content = remove_whitespace(doc.page_content)
        big_doc.page_content += f"\n{content}"
        
        image_gsurl = upload_doc_images(doc.metadata)
        if image_gsurl:
            doc.metadata["image_gsurl"] = image_gsurl
            doc.metadata["image_base64"] = None
            doc.metadata["uploaded_to_bucket"] = True
            big_doc.metadata["images_gsurls"].append(image_gsurl)

        for key, value in doc.metadata.items():
            if key not in big_doc.metadata:
                big_doc.metadata[key] = value
        
        big_doc.metadata["chunk_metadata"].append(doc.metadata)

    big_doc.metadata["doc_id"] = doc_id
    big_doc.metadata["char_count"] = len(big_doc.page_content)

    if len(big_doc.page_content) == 0 and not big_doc.metadata.get("images_gsurls"):
        log.warning("No content found to add for big_doc")
        return None, None, docs
    
    # Serialize lists in metadata to JSON strings before saving
    for key in ["images_gsurls", "chunk_metadata"]:
        big_doc.metadata[key] = json.dumps(big_doc.metadata[key])

    source = big_doc.metadata.get("source") or doc.metadata.get("url")
    if not source:
        log.warning(f"No source found for big_doc {doc_id} {big_doc.metadata}")
    
    return doc_id, big_doc, docs

def summarise_docs(docs, vector_name, summary_threshold_default=10000, model_limit_default=25000):

    if not docs:
        return None

    chunker_config = load_config_key("chunker", vector_name=vector_name, kind="vacConfig")
    summarise_chunking_config = chunker_config.get("summarise") if chunker_config else None
    
    if not summarise_chunking_config:
        return None

    # if model not specified will use default config.llm
    model = summarise_chunking_config.get("model")
    summary_llm_str   = summarise_chunking_config.get("llm") if summarise_chunking_config.get("llm") else chunker_config.get("llm")
    summary_threshold = summarise_chunking_config.get("threshold") if summarise_chunking_config.get("threshold") else summary_threshold_default
    model_limit = summarise_chunking_config.get("model_limit") if summarise_chunking_config.get("model_limit") else model_limit_default

    summary_llm = llm_str_to_llm(summary_llm_str, model=model)

    doc_summaries = {}
    for doc in docs:
        try:
            if len(doc.page_content) > summary_threshold:

                context = doc.page_content[:model_limit]
                metadata = doc.metadata
                if len(doc.page_content) > model_limit:
                    log.warning(f"Page content was above model_limit for summary: [{len(doc.page_content)} / {model_limit}]: {metadata}")
                    #TODO: use map_reduce chain for summary instead

                log.info(f"Creating summary for {metadata} for doc [{len(context)}]")
                
                prompt_template = """
Summarise the context below.  
Include in the summary the title showing which document it comes from, and ensure you include any companies, people or entites involved within the text. 
Finish the summary with a list of relevant keywords that will help the summary be found via search engines. 
Be careful not to add any speculation or any details that are not covered in the original:
## Context:
{context}
## Metadata
{metadata}
## Your Summary:
"""                
                prompt = PromptTemplate.from_template(prompt_template)
                summary_chain = prompt | summary_llm | StrOutputParser()
                doc_source = doc.metadata.get('source')
                doc_objectId = doc.metadata.get('objectId')
                doc_eventTime = doc.metadata.get('eventTime')

                full_context = f"# {doc_source}\n\n## {doc_objectId}\n{doc_eventTime}\n\n## Text\n{context}"

                str_metadata = json.dumps(metadata)

                cb_langfuse = create_langfuse_callback()

                summary = summary_chain.invoke(
                        {"context": full_context, "metadata": str_metadata[:2000]},
                        config={"callbacks": [cb_langfuse]}
                    )
                
                log.info(f"Created a summary for {metadata}: {len(context)} > {len(summary)}")
                
                if len(summary) < 100:
                    log.info(f"Summary not long enough {metadata}, skipping")
                    continue

                # Create a temporary file for the summary
                bucket_name = os.getenv("DOC_BUCKET")
                if not bucket_name:
                    raise ValueError("No DOC_BUCKET configured for summary")
                with tempfile.NamedTemporaryFile(mode='w+', delete=False) as temp_file:
                    temp_file.write(summary)
                    temp_file.flush()
                    temp_file_path = temp_file.name

                    log.info(f"Wrote summary to file {temp_file_path}: {summary[30:]}")

                    file_size = os.path.getsize(temp_file_path)
                    if file_size > 0:
                        bucket_filepath = get_summary_file_name(metadata["objectId"])
                        summary_loc = add_file_to_gcs(temp_file_path, 
                                                    vector_name=vector_name, 
                                                    bucket_name=bucket_name, 
                                                    metadata=metadata, 
                                                    bucket_filepath=bucket_filepath)
                        
                        doc_summaries[metadata["objectId"]] = summary_loc
                    else:
                        log.error(f"Summary file {temp_file_path} is empty. Skipping upload.")
        except Exception as err:
            log.error(f"Failed to create a summary for {metadata}: {str(err)} traceback: {traceback.format_exc()}")
    
    
    return doc_summaries




import pytest
from unittest.mock import patch
from sunholo.agents.dispatch_to_qa import prep_request_payload

@pytest.mark.parametrize("user_input, chat_history, vector_name, stream, expected_endpoint, expected_payload", [
    ("What is AI?", [], "my_vector", False, "http://example.com/invoke", {"user_input": "What is AI?", "chat_history": [], "vector_name": "my_vector"}),
    ("Tell me about ML", ["Previous chat message"], "another_vector", True, "http://example.com/stream", {"user_input": "Tell me about ML", "chat_history": ["Previous chat message"], "vector_name": "another_vector"}),
    # Additional test cases to cover all logic paths
    ("", [], "empty_input_vector", False, "http://example.com/invoke", {"user_input": "", "chat_history": [], "vector_name": "empty_input_vector"}),
    ("Valid input", ["Chat history present"], "history_vector", False, "http://example.com/invoke", {"user_input": "Valid input", "chat_history": ["Chat history present"], "vector_name": "history_vector"}),
    ("Stream request", [], "stream_vector", True, "http://example.com/stream", {"user_input": "Stream request", "chat_history": [], "vector_name": "stream_vector"}),
    # New test cases for prep_request_payload function
    ("User input with override endpoint", [], "override_vector", False, "http://override.com/invoke", {"user_input": "User input with override endpoint", "chat_history": [], "vector_name": "override_vector"}),
    ("Stream with chat history and override endpoint", ["Chat history for override"], "override_stream_vector", True, "http://override.com/stream", {"user_input": "Stream with chat history and override endpoint", "chat_history": ["Chat history for override"], "vector_name": "override_stream_vector"})
])
@patch('sunholo.agents.dispatch_to_qa.load_config_key')
@patch('sunholo.agents.dispatch_to_qa.route_endpoint')
def test_prep_request_payload(mock_route_endpoint, mock_load_config_key, user_input, chat_history, vector_name, stream, expected_endpoint, expected_payload):
    mock_route_endpoint.return_value = {"invoke": "http://example.com/invoke", "stream": "http://example.com/stream"}
    mock_load_config_key.side_effect = lambda key, vector_name=None, kind=None: "my_vector" if key == "vector_name" else None
    # Mocking for override endpoint scenario
    if vector_name.startswith("override"):
        mock_route_endpoint.return_value = {"invoke": "http://override.com/invoke", "stream": "http://override.com/stream"}

    endpoint, payload = prep_request_payload(user_input, chat_history, vector_name, stream)

    assert endpoint == expected_endpoint
    assert payload == expected_payload



try:
    from vertexai.preview import rag
except ImportError:
    rag = None

from ..logging import log
from ..utils.config import load_config_key
from ..vertex import init_vertex
from .get_files import fetch_corpus
from ..components import load_memories
    

def do_llamaindex(message_data, metadata, vector_name):
    """
    Configures and manages the corpus for a VertexAI project using the specified vector name
    by importing message data from Google Cloud Storage or Google Drive URLs.

    This function loads configuration from a YAML file, initializes a Vertex AI environment,
    and either fetches an existing corpus or creates a new one if it doesn't exist.
    It supports importing files directly from cloud storage links.

    Parameters:
        message_data (str): The URL to the data on Google Cloud Storage or Google Drive that needs to be imported to the corpus.
        metadata (dict): Additional metadata not explicitly used in this function but might be needed for extended functionality.
        vector_name (str): The name of the vector (and corpus) which will be used to locate and configure the specific settings from the configuration files.

    Raises:
        ValueError: If the necessary configurations for GCP or project ID are not found, or if the corpus could not be established.
        NotImplementedError: If the data is not from supported sources (Google Cloud Storage or Google Drive).

    Example:
    ```python
    message_data = "gs://bucket_name/path_to_file.txt"
    metadata = {"user": "admin"}
    vector_name = "example_vector"
    response = do_llamaindex(message_data, metadata, vector_name)
    print(response)
    # Imported file to corpus: {'status': 'success'}
    ```
    """
    if not rag:
        raise ValueError("Need to install vertexai module via `pip install sunholo[gcp]`")

    gcp_config = load_config_key("gcp_config", vector_name=vector_name, kind="vacConfig")
    if not gcp_config:
        raise ValueError(f"Need config.{vector_name}.gcp_config to configure llamaindex on VertexAI")

    init_vertex(gcp_config)

    global_project_id = gcp_config.get('project_id')
    global_location = gcp_config.get('location')
    global_rag_id = gcp_config.get('rag_id')
    #global_data_store_id = gcp_config.get('data_store_id')

    memories = load_memories(vector_name)
    tools = []

    if not memories:
        return tools
    
    corpuses = []
    for memory in memories:
        for key, value in memory.items():  # Now iterate over the dictionary
            vectorstore = value.get('vectorstore')
            if vectorstore == "llamaindex":
                log.info(f"Found vectorstore {vectorstore}")
                if value.get('read_only'):
                    continue
                rag_id = value.get('rag_id')
                project_id = gcp_config.get('project_id')
                location = gcp_config.get('location')
                corpus = fetch_corpus(
                    project_id=project_id or global_project_id,
                    location=location or global_location,
                    rag_id=rag_id or global_rag_id
                )
                corpuses.append(corpus)
    if not corpuses:
        log.error("Could not find a RAG corpus to import data to")
        return None
    
    try:
        corpura = rag.list_corpora()
        log.info(f"All Corpora: {corpura} - {type(corpura)}")
    except Exception as err:
        log.warning(f"Could not list any corpora - {str(err)}")

    log.info(f"Found llamaindex corpus: {corpuses}")

    # native support for cloud storage and drive links
    chunker_config = load_config_key("chunker", vector_name=vector_name, kind="vacConfig")

    
    if message_data.startswith("gs://") or message_data.startswith("https://drive.google.com"):
        log.info(f"rag.import_files for {message_data}")
        for corp in corpuses:
            response = rag.import_files(
                corpus_name=corp.name,
                paths=[message_data],
                chunk_size=chunker_config.get("chunk_size"),  # Optional
                chunk_overlap=chunker_config.get("overlap"),  # Optional
            )
            log.info(f"Imported file to corpus: {response} with metadata: {metadata}")

        metadata["source"] = message_data
        return metadata
    
    else:
        log.warning("Only gs:// and https://drive data is supported for llamaindex")
        

def check_llamaindex_in_memory(vector_name):
    memories = load_config_key("memory", vector_name=vector_name, kind="vacConfig")
    for memory in memories:  # Iterate over the list
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore:
                log.info(f"Found vectorstore {vectorstore}")
                if vectorstore == "llamaindex":

                    return True
    
    return False

def llamaindex_chunker_check(message_data, metadata, vector_name):
    # llamaindex handles its own chunking/embedding
    memories = load_config_key("memory", vector_name=vector_name, kind="vacConfig")
    total_memories = len(memories)
    llama = None
    if check_llamaindex_in_memory(vector_name):
        llama = do_llamaindex(message_data, metadata, vector_name)
        log.info(f"Processed llamaindex: {llama}")

    # If llamaindex is the only entry, return
    if llama and total_memories == 1:

        return llama
    
    elif llama:
        log.info("Llamaindex found but not the only memory, continuing with other processes.")

        return None


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import log
from .vectorstore import pick_vectorstore
from ..utils import load_config_key
from .llm import get_embeddings
from ..utils.gcp_project import get_gcp_project

from langchain.retrievers import MergerRetriever
from langchain_community.retrievers import GoogleCloudEnterpriseSearchRetriever
# https://python.langchain.com/docs/integrations/retrievers/merger_retriever
from langchain_community.document_transformers import EmbeddingsRedundantFilter
from langchain.retrievers.document_compressors import DocumentCompressorPipeline
from langchain.retrievers import ContextualCompressionRetriever



def load_memories(vector_name):
    memories = load_config_key("memory", vector_name, kind="vacConfig")
    log.info(f"Found memory settings for {vector_name}: {memories}")
    if not memories or len(memories) == 0:
        log.info(f"No memory settings found for {vector_name}")
        return None

    return memories

def pick_retriever(vector_name, embeddings=None):

    memories = load_memories(vector_name)

    retriever_list = []
    for memory in memories:  # Iterate over the list
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore:
                log.info(f"Found vectorstore {vectorstore}")
                from_metadata_id = value.get('from_metadata_id')
                if from_metadata_id:
                    # this entry needs to be fetched via metadata key
                    log.info(f"Skipped from_metadata_id for {vectorstore}")
                    continue

                embeddings = embeddings or get_embeddings(vector_name)
                read_only = value.get('read_only')
                try:
                    vectorstore = pick_vectorstore(vectorstore, 
                                                vector_name=vector_name, 
                                                embeddings=embeddings, 
                                                read_only=read_only)
                except Exception as e:
                    log.error(f"Failed to pick_vectorstore {vectorstore} for {vector_name} - {str(e)} - skipping")
                    continue
                
                k_override = value.get('k', 3)
                vs_retriever = vectorstore.as_retriever(search_kwargs=dict(k=k_override))
                retriever_list.append(vs_retriever)
            
            if value.get('provider') == "GoogleCloudEnterpriseSearchRetriever":
                log.info(f"Found GoogleCloudEnterpriseSearchRetriever {value['provider']}")
                gcp_retriever = GoogleCloudEnterpriseSearchRetriever(
                    project_id=get_gcp_project(),
                    search_engine_id=value["db_id"],
                    location_id=value.get("location", "global"),
                    engine_data_type=1 if value.get("type","unstructured") == "structured" else 0,
                    query_expansion_condition=2
                )
                retriever_list.append(gcp_retriever)
            
            #TODO: more memory stores here

    if not retriever_list or len(retriever_list) == 0:
        log.info(f"No retrievers were created for {memories}")
        return None
    
    retriever = process_retrieval(retriever_list, vector_name)

    return retriever

def metadata_retriever(metadata: dict, key: str, vector_name:str, embeddings=None):
    """
    Decides which vector_name to retrieve from metadata passed
    """
    memories = load_memories(vector_name)

    retriever_list = []
    for memory in memories:  # Iterate over the list
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore:
                log.info(f"Found vectorstore {vectorstore}")
                from_metadata_id = value.get('from_metadata_id')
                if from_metadata_id:
                    # this entry needs to be fetched via metadata key
                    log.info(f"Finding id from_metadata_id for {vectorstore}")
                    if key not in metadata:
                        raise ValueError(f"Missing {key} in {metadata}")
                    the_id = metadata[key]
                    read_only = value.get('read_only')
                    embeddings = embeddings or get_embeddings(vector_name)
                    vectorstore = pick_vectorstore(vectorstore, 
                                                   vector_name=the_id, 
                                                   embeddings=embeddings, 
                                                   read_only=read_only)
                    k_override = value.get('k', 3)
                    id_retriever = vectorstore.as_retriever(search_kwargs=dict(k=k_override))
                    retriever_list.append(id_retriever)
                else:
                    continue

    if not retriever_list or len(retriever_list) == 0:
        log.info(f"No retrievers were created for {memories}")
        return None
    
    retriever = process_retrieval(retriever_list, vector_name)

    return retriever
    


def process_retrieval(retriever_list: list, vector_name: str):
    k_override = load_config_key("memory_k", vector_name, kind="vacConfig")
    lotr = MergerRetriever(retrievers=retriever_list)

    filter_embeddings = get_embeddings(vector_name)
    filter = EmbeddingsRedundantFilter(embeddings=filter_embeddings)
    pipeline = DocumentCompressorPipeline(transformers=[filter])
    retriever = ContextualCompressionRetriever(
        base_compressor=pipeline, base_retriever=lotr, 
        k=k_override)
    
    return retriever